
/*=============================================================================
   rtf2htm, a command-line program to convert RTF-formatted documents to HTML.
   Copyright (C) 2000,2001 Zachary Thayer Smith

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   The author is reachable by electronic mail at tuorfa@yahoo.com.
=============================================================================*/


/*----------------------------------------------------------------------
 * Module name:    convert
 * Author name:    Zach Smith
 * Create date:    24 Jul 2001
 * Last change:    03 Aug 2001
 * Purpose:        Performs conversion from RTF to HTML.
 *----------------------------------------------------------------------
 * Changes:
 * 24 Jul 01, tuorfa@yahoo.com: moved code over from word.c
 * 24 Jul 01, tuorfa@yahoo.com: fixed color table reference numbering.
 * 30 Jul 01, tuorfa@yahoo.com: moved hex convert to util.c
 * 30 Jul 01, tuorfa@yahoo.com: moved special expr tables to special.c
 * 30 Jul 01, tuorfa@yahoo.com: moved attribute stack to attr.c
 * 31 Jul 01, tuorfa@yahoo.com: began addition of hash of rtf commands
 * 01 Aug 01, tuorfa@yahoo.com: finished bulk of rtf command hash
 * 03 Aug 01, tuorfa@yahoo.com: removed no-op hash entries to save space
 * 03 Aug 01, tuorfa@yahoo.com: code to ignore rest of groups for \*, etc
 * 03 Aug 01, tuorfa@yahoo.com: fixed para-alignnot being cleared by \pard
 * 03 Aug 01, tuorfa@yahoo.com: added support for \keywords group
 * 03 Aug 01, tuorfa@yahoo.com: added dummy funcs for header/footer
 * 03 Aug 01, tuorfa@yahoo.com: began addition of hyperlink support
 * 04 Aug 01, tuorfa@yahoo.com: fixed debug string printing
 * 05 Aug 01, tuorfa@yahoo.com: added support for hyperlink data with \field
 * 06 Aug 01, tuorfa@yahoo.com: added support for several font attributes
 *--------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#include "defs.h"
#include "parse.h"
#include "util.h"
#include "malloc.h"
#include "main.h"
#include "error.h"
#include "word.h"
#include "hash.h"
#include "special.h"
#include "attr.h"


static int charset_type=CHARSET_ANSI;


/* Nested tables aren't supported.
 * Notice that <th> isn't here, since RTF doesn't have a direct equivalent.
 */
static int coming_pars_that_are_tabular = FALSE;
static int within_table = FALSE;
static int have_printed_tr=FALSE;
static int have_printed_td=FALSE;
static int have_printed_end_tr=FALSE;
static int have_printed_end_td=FALSE;


/* Previously in word_print_html_core function 
 */
static int total_chars_this_line=0; /* for simulating \tab */


/* Paragraph alignment (kludge)
 */
enum {
	ALIGN_LEFT=0,
	ALIGN_RIGHT,
	ALIGN_CENTER,
	ALIGN_JUSTIFY
};



/* This value is set by attr_push and attr_pop 
 */
int simulate_smallcaps;
int simulate_allcaps;


/* Most pictures must be written to files. */
enum {
	PICT_UNKNOWN=0,
	PICT_WM,
	PICT_MAC,
	PICT_PM,
	PICT_DI,
	PICT_WB
};
static int within_picture=FALSE;
static int picture_file_number=1;
static char picture_path[255];
static int picture_width;
static int picture_height;
static int picture_bits_per_pixel=1;
static int picture_type=PICT_UNKNOWN;
static int picture_wmetafile_type;
static char *picture_wmetafile_type_str;


static int have_printed_body=FALSE;
static int within_head=TRUE;



static char *hyperlink_base = NULL;



void starting_body();
void starting_text();


static int banner_printed=FALSE;

void
print_banner () {
	if (!banner_printed) {
	  printf ("<!-- Translation from RTF performed by rtf2htm version %s\n",
		PROGRAM_VERSION);
	  printf (" !-- For information about this marvellous Open Source program\n");
	  printf (" !-- please go to http://recycle.50megs.com\n !-->\n");
	}
	banner_printed=TRUE;
}


void
starting_body ()
{
	if (!have_printed_body) {
		if (!inline_mode)
			printf ("</HEAD><BODY>\n");
		within_head=FALSE;
		have_printed_body=TRUE;
	}
}


/*-------------------------------------------------------------------*/
/*-------------------------------------------------------------------*/
/*-------------------------------------------------------------------*/

/* Hardly an important group, since the data contained is often
 * inaccurate, but we may as well print it out.
 */

static char *month_strings[12]= {
#ifdef ENGLISH
  "January","February","March","April","May","June","July","August",
  "September","October","November","December"
#endif
#ifdef FRANCAIS
  "Janvier","Fevrier","Mars","Avril","Mai","Juin","Juillet","Aout","Septembre",
  "Octobre","Novembre","Decembre"
#endif
#ifdef ITALIANO
  "Ianuario","Febbraio","Marzo","Aprile","Maggio","Iuno",
  "Luglio","Agusto","Settembre","Ottobre","Novembre","Dicembre"
#endif
#ifdef ESPANOL
  "?","Febraio","Marzo","Abril","Mayo","?","?","Agosto",
  "Septiembre","Octubre","Noviembre","Diciembre"
#endif
#ifdef DEUTCH
#endif
};

void word_dump_date(Word *w)
{
	int year=0, month=0, day=0, hour=0, minute=0;
	CHECK_PARAM_NOT_NULL(w);
	while (w) {
	 	char *s = word_string (w);
		if (*s == '\\') {
			++s;
			if (!strncmp (s, "yr", 2) && isdigit(s[2])) {
				year = atoi (&s[2]);
			}
			else if (!strncmp (s, "mo", 2) && isdigit(s[2])) {
				month= atoi (&s[2]);
			}
			else if (!strncmp (s, "dy", 2) && isdigit(s[2])) {
				day= atoi (&s[2]);
			}
			else if (!strncmp (s, "min", 3) && isdigit(s[3])) {
				minute= atoi (&s[3]);
			}
			else if (!strncmp (s, "hr", 2) && isdigit(s[2])) {
				hour= atoi (&s[2]);
			}
		}
		w=w->next;
	}
	if (year && month && day) {
		printf ("%d %s %d ", day, month_strings[month-1], year);
	}
	if (hour && minute) {
		printf ("%02d:%02d ", hour, minute);
	}
}



/*-------------------------------------------------------------------*/

typedef struct {
	int num;
	char *name;
} FontEntry;

#define MAX_FONTS (256)
static FontEntry font_table[MAX_FONTS];
static int total_fonts=0;

char* lookup_fontname (int num) {
	int i;
	if (total_fonts) 
	for(i=0;i<total_fonts;i++) {
		if (font_table[i].num==num) 
			return font_table[i].name;
	}
	return NULL;
}

void
process_font_table (Word *w)
{
	Word *w2;

	CHECK_PARAM_NOT_NULL(w);

	while(w) {
		int num;
		char name[255];
		char *tmp;

		if ((w2=w->child)) {
			tmp = word_string (w2);
			if (!strncmp("\\f",tmp,2)) {
				num = atoi (&tmp[2]);
				name[0]=0;

				w2=w2->next;
				while(w2) {
					tmp = word_string (w2);
					if (tmp && tmp[0] != '\\')
						strcat(name,tmp);

					w2=w2->next;
				}

				/* Chop the gall-derned semicolon. */
				if ((tmp=strchr(name,';')))
					*tmp=0;

				font_table[total_fonts].num=num;
				font_table[total_fonts].name=my_strdup(name);
				total_fonts++;
			}
		}
		w=w->next;
	}

	printf ("<!-- font table contains %d fonts total-->\n",total_fonts);

	if (debug_mode) {
		int i;

		printf ("<!-- font table dump: \n");
		for (i=0; i<total_fonts; i++) {
			printf (" font %d = %s\n", font_table[i].num, 
				font_table[i].name);
		}
		printf ("-->\n");
	}
}


void
process_info_group (Word *w)
{
	Word *child;

	CHECK_PARAM_NOT_NULL(w);

	while(w) {
		child = w->child;
		if (child) {
			Word *w2;
			char *s;

			s = word_string(child);

			if (!inline_mode) {
				if (!strcmp("\\title", s)) {
					printf ("<title>");
					w2=child->next;
					while (w2) {
						char *s2 = word_string(w2);
						if (s2[0] != '\\') printf ("%s ", s2);
						w2=w2->next;
					}
					printf ("</title>\n");
				}
				else if (!strcmp("\\keywords", s)) {
					printf ("<meta name=keywords value=\"");
					w2=child->next;
					while (w2) {
						char *s2 = word_string(w2);
						if (s2[0] != '\\') printf ("%s,", s2);
						w2=w2->next;
					}
					printf ("\">\n");
				}
				else if (!strcmp("\\author", s)) {
					printf ("<!-- author: ");
					w2=child->next;
					while (w2) {
						char *s2 = word_string(w2);
						if (s2[0] != '\\') printf ("%s ", s2);
						w2=w2->next;
					}
					printf (" -->\n");
				}
				else if (!strcmp("\\comment", s)) {
					printf ("<!-- comment: ");
					w2=child->next;
					while (w2) {
						char *s2 = word_string(w2);
						if (s2[0] != '\\') printf ("%s ", s2);
						w2=w2->next;
					}
					printf (" -->\n");
				}
				else if (!strncmp("\\nofpages", s, 9)) {
					printf ("<!-- total pages: %s -->\n",&s[9]);
				}
				else if (!strncmp("\\nofwords", s, 9)) {
					printf ("<!-- total words: %s -->\n",&s[9]);
				}
				else if (!strncmp("\\nofchars", s, 9) && isdigit(s[9])) {
					printf ("<!-- total chars: %s -->\n",&s[9]);
				}
				else if (!strncmp("\\edmins", s, 7)) {
					printf ("<!-- total editing minutes %s -->\n",
						&s[7]);
				}
				else if (!strcmp("\\creatim", s)) {
					printf ("<!-- creaton date: ");
					if (child->next) word_dump_date (child->next);
					printf (" -->\n");
				}
				else if (!strcmp("\\printim", s)) {
					printf ("<!-- last printed: ");
					if (child->next) word_dump_date (child->next);
					printf (" -->\n");
				}
				else if (!strcmp("\\buptim", s)) {
					printf ("<!-- last backup: ");
					if (child->next) word_dump_date (child->next);
					printf (" -->\n");
				}
				else if (!strcmp("\\revtim", s)) {
					printf ("<!-- revision date: ");
					if (child->next) word_dump_date (child->next);
					printf (" -->\n");
				}
			}

			/* Irregardless of whether we're in inline mode,
			 * we want to process the following.
			 */
			if (!strcmp("\\hlinkbase", s)) {
				char *linkstr = NULL;

				printf ("<!-- hyperlink base: ");
				if (child->next) {
					Word *nextword = child->next;

					if (nextword) 
						linkstr=word_string (nextword);
				}

				if (linkstr)
					printf ("%s", linkstr);
				else
					printf ("(none)");
				printf (" -->\n");

				/* Store the pointer, it will remain good. */
				hyperlink_base = linkstr; 
			}
		} 
		w = w->next;
	}
}

/*-------------------------------------------------------------------*/

typedef struct {
	unsigned char r,g,b;
} Color;

#define MAX_COLORS (256)
static Color color_table[MAX_COLORS];
static int total_colors=0;

void
process_color_table (Word *w)
{
	int r,g,b;

	CHECK_PARAM_NOT_NULL(w);

	/* Sometimes, RTF color tables begin with a semicolon,
	 * i.e. an empty color entry. This seems to indicate that color 0 
	 * will not be used, so here I set it to black.
	 */
	r=g=b=0;

	while(w) {
		char *s = word_string (w);

#if 0
		printf ("<!- found this color table word: %s -->\n", word_string(w));
#endif

		if(!strncmp("\\red",s,4)) {
			r = atoi(&s[4]);
			while(r>255) r>>=8;
		}
		else if(!strncmp("\\green",s,6)) {
			g = atoi(&s[6]);
			while(g>255) g>>=8;
		}
		else if(!strncmp("\\blue",s,5)) {
			b = atoi(&s[5]);
			while(b>255) b>>=8;
		}
		else
		/* If we find the semicolon which denotes the end of
		 * a color entry then store the color, even if we don't
		 * have all of it.
		 */
		if (!strcmp (";", s)) {
			color_table[total_colors].r = r;
			color_table[total_colors].g = g;
			color_table[total_colors++].b = b;
			if (debug_mode) 
				printf ("<!-- storing color entry %d: 0x%02x%02x%02x -->\n", total_colors-1, r,g,b);
			r=g=b=0;
		}

		w=w->next;
	}

	if (debug_mode)
	  printf ("<!-- color table had %d entries -->\n", total_colors);
}

/*-------------------------------------------------------------------*/

static int cmd_cf (Word *w, int align, char has_param, short num) {
	char str[40];

	if (!has_param || num>=total_colors) {
		printf ("<! -- font color change attempted which is invalid %d -- >\n", num);
	}
	else
	{
		sprintf (str,"#%02x%02x%02x",
			color_table[num].r,
			color_table[num].g,
			color_table[num].b);
		attr_push(ATTR_FOREGROUND,str);
	}
	return FALSE;
}

static int cmd_cb (Word *w, int align, char has_param, short num) {
	char str[40];

	if (!has_param || num>=total_colors) {
		printf ("<! -- font color change attempted which is invalid %d -- >\n", num);
	}
	else
	{
		sprintf (str,"#%02x%02x%02x",
			color_table[num].r,
			color_table[num].g,
			color_table[num].b);
		attr_push(ATTR_BACKGROUND,str);
	}
	return FALSE;
}

static int cmd_fs (Word *w, int align, char has_param, short points) {
	char str[20];

	if (!has_param) return FALSE;

	/* Note, fs20 means 10pt */
	points /= 2;

	sprintf (str,"%d",points);
	strcat (str,"pt");
	switch (points) {
	  case 8:
	    attr_push(ATTR_STD_FONTSIZE,"1");
	    break;
	  case 10:
	    attr_push(ATTR_STD_FONTSIZE,"2");
	    break;
	  case 12:
	    attr_push(ATTR_STD_FONTSIZE,"3");
	    break;
	  case 14:
	    attr_push(ATTR_STD_FONTSIZE,"4");
	    break;
	  case 18:
	    attr_push(ATTR_STD_FONTSIZE,"5");
	    break;
	  case 24:
	    attr_push(ATTR_STD_FONTSIZE,"6");
	    break;

	  default:
	    if (!simple_mode) {
	      attr_push(ATTR_FONTSIZE,str);
	    } else {
		if (points <= 8)
		  attr_push(ATTR_STD_FONTSIZE,"1");
		else if (points <= 10)
		  attr_push(ATTR_STD_FONTSIZE,"2");
		else if (points <= 12)
		  attr_push(ATTR_STD_FONTSIZE,"3");
		else if (points <= 14)
		  attr_push(ATTR_STD_FONTSIZE,"4");
		else if (points <= 18)
		  attr_push(ATTR_STD_FONTSIZE,"5");
		else if (points <= 24)
		  attr_push(ATTR_STD_FONTSIZE,"6");
		else 
		  attr_push(ATTR_STD_FONTSIZE,"7");
	    }
	}
	return FALSE;
}


/* Name:     cmd_field
 * Purpose:  Interprets fields looking for hyperlinks.
 * Comment:  Because hyperlinks are put in \field groups,
 *           we must interpret all \field groups, which is
 *           slow and laborious.
 */

static int cmd_field (Word *w, int align, char has_param, short num) {
	Word *child;

	CHECK_PARAM_NOT_NULL(w);

	while(w) {
		child = w->child;
		if (child) {
			Word *w2;
			char *s;

			s = word_string(child);

			if (!strcmp("\\*", s)) {
				w2=child->next;
				while (w2) {
					char *s2 = word_string(w2);
					if (!strcmp("\\fldinst", s2)) {
						Word *w3;
						w3=w2->next;
						while (w3 && !w3->child) {
							w3=w3->next;
						}
						if (w3) w3=w3->child;
						while (w3) {
							char *s3=word_string(w3);
							if (!strcmp("HYPERLINK",s3)) {
								Word *w4;
								char *s4;
								w4=w3->next;
								while (w4 && !strcmp(" ", word_string(w4)))
									w4=w4->next;
								if (w4) {
									s4=word_string(w4);
									printf ("<a href=%s>%s</a>\n", s4,s4);
									return TRUE;
								}
								
							}
							w3=w3->next;
						}
					}
					w2=w2->next;
				}
				
			}
		}
		w=w->next;
	}
	return TRUE;
}

static int cmd_f (Word *w, int align, char has_param, short num) {
	char *name;

	/* no param exit early XX */
	if (!has_param) 
		return FALSE;

	name = lookup_fontname(num);
	if(!name) {
		printf ("<!-- invalid font number %d -->\n",num);
	} else {
		attr_push(ATTR_FONTFACE,name);
	}

	return FALSE;
}

static int cmd_highlight (Word *w, int align, char has_param, short num) {
	char str[40];

	if (!has_param || num>=total_colors) {
		printf ("<! -- font background color change attempted which is invalid %d -- >\n", num);
	}
	else
	{
		sprintf (str,"#%02x%02x%02x",
			color_table[num].r,
			color_table[num].g,
			color_table[num].b);
		attr_push(ATTR_BACKGROUND,str);
	}
	return FALSE;
}

static int cmd_tab (Word *w, int align, char has_param, short param) {

	/* Tab presents a genuine problem
	 * since HTML doesn't have an equivalent.
	 * As a kludge fix, I shall assume
	 * the font is fixed width and that
	 * the tabstops are 8 characters apart.
	 */
	int need= 8-(total_chars_this_line%8);
	total_chars_this_line += need;
	while(need>0) {
		printf ("&nbsp;");
		need--;
	}
	printf ("\n");
	return FALSE;
}

static int cmd_plain (Word *w, int align, char has_param, short param) {
	attr_pop_all();
	return FALSE;
}
static int cmd_fnil (Word *w, int align, char has_param, short param) {
	attr_push(ATTR_FONTFACE,FONTNIL_STR);
	return FALSE;
}
static int cmd_froman (Word *w, int align, char has_param, short param) {
	attr_push(ATTR_FONTFACE,FONTROMAN_STR);
	return FALSE;
}
static int cmd_fswiss (Word *w, int align, char has_param, short param) {
	attr_push(ATTR_FONTFACE,FONTSWISS_STR);
	return FALSE;
}
static int cmd_fmodern (Word *w, int align, char has_param, short param) {
	attr_push(ATTR_FONTFACE,FONTMODERN_STR);
	return FALSE;
}
static int cmd_fscript (Word *w, int align, char has_param, short param) {
	attr_push(ATTR_FONTFACE,FONTSCRIPT_STR);
	return FALSE;
}
static int cmd_fdecor (Word *w, int align, char has_param, short param) {
	attr_push(ATTR_FONTFACE,FONTDECOR_STR);
	return FALSE;
}
static int cmd_ftech (Word *w, int align, char has_param, short param) {
	attr_push(ATTR_FONTFACE,FONTTECH_STR);
	return FALSE;
}

static int cmd_expand (Word *w, int align, char has_param, short param) {
	char str[10];
	if (has_param) {
		sprintf (str, "%d", param/4);
		if (!param) 
			attr_pop(ATTR_EXPAND);
		else 
			attr_push(ATTR_EXPAND, str);
	}
	return FALSE;
}

static int cmd_emboss (Word *w, int align, char has_param, short param) {
	char str[10];
	if (has_param && !param)
		attr_pop(ATTR_EMBOSS);
	else
	{
		sprintf (str, "%d", param);
		attr_push(ATTR_EMBOSS, str);
	}
	return FALSE;
}

static int cmd_engrave (Word *w, int align, char has_param, short param) {
	char str[10];
	if (has_param && !param) 
		attr_pop(ATTR_ENGRAVE);
	else
	{
		sprintf (str, "%d", param);
		attr_push(ATTR_ENGRAVE, str);
	}
	return FALSE;
}

static int cmd_caps (Word *w, int align, char has_param, short param) {
	if (has_param && !param)
		attr_pop(ATTR_CAPS);
	else 
		attr_push(ATTR_CAPS,NULL);
	return FALSE;
}

static int cmd_scaps (Word *w, int align, char has_param, short param) {
	if (has_param && !param)
		attr_pop(ATTR_SMALLCAPS);
	else 
		attr_push(ATTR_SMALLCAPS,NULL);
	return FALSE;
}

static int cmd_bullet (Word *w, int align, char has_param, short param) {
	printf ("&bull;");
	++total_chars_this_line; /* \tab */
	return FALSE;
}

static int cmd_ldblquote (Word *w, int align, char has_param, short param) {
	printf ("&ldquo;");
	++total_chars_this_line; /* \tab */
	return FALSE;
}

static int cmd_rdblquote (Word *w, int align, char has_param, short param) {
	printf ("&rdquo;");
	++total_chars_this_line; /* \tab */
	return FALSE;
}

static int cmd_lquote (Word *w, int align, char has_param, short param) {
	printf ("&lsquo;");
	++total_chars_this_line; /* \tab */
	return FALSE;
}

static int cmd_rquote (Word *w, int align, char has_param, short param) {
	printf ("&rsquo;");
	++total_chars_this_line; /* \tab */
	return FALSE;
}

static int cmd_par (Word *w, int align, char has_param, short param) {
	printf ("<br>");
	total_chars_this_line = 0; /* \tab */
	return FALSE;
}
static int cmd_line (Word *w, int align, char has_param, short param) {
	printf ("<br>\n");
	total_chars_this_line = 0; /* \tab */
	return FALSE;
}
static int cmd_page (Word *w, int align, char has_param, short param) {
	printf ("<hr><hr>\n");
	total_chars_this_line = 0; /* \tab */
	return FALSE;
}
static int cmd_intbl (Word *w, int align, char has_param, short param) {
	++coming_pars_that_are_tabular;
	return FALSE;
}

static int cmd_ulnone (Word *w, int align, char has_param, short param) {
	int attr, more=TRUE;
	do {
		attr = attr_read();
		if (attr==ATTR_UNDERLINE ||
		    attr==ATTR_DOT_UL ||
		    attr==ATTR_DASH_UL ||
		    attr==ATTR_DOT_DASH_UL ||
		    attr==ATTR_2DOT_DASH_UL ||
		    attr==ATTR_WORD_UL ||
		    attr==ATTR_WAVE_UL ||
		    attr==ATTR_THICK_UL ||
		    attr==ATTR_DOUBLE_UL)
		{
		  if (!attr_pop(ATTR_UNDERLINE))
		    ;
		} else
		  more=FALSE;
	} while(more);
	return FALSE;
}

static int cmd_ul (Word *w, int align, char has_param, short param) {
	if (has_param && param != 0) {
		attr_push(ATTR_UNDERLINE,NULL);
	} else {
		cmd_ulnone (w, align, has_param, param);
	}
	return FALSE;
}

static int cmd_uld (Word *w, int align, char has_param, short param) {
	attr_push(ATTR_DOUBLE_UL,NULL);
	return FALSE;
}

static int cmd_uldb (Word *w, int align, char has_param, short param) {
	attr_push(ATTR_DOT_UL,NULL);
	return FALSE;
}

static int cmd_uldash (Word *w, int align, char has_param, short param) {
	attr_push(ATTR_DASH_UL,NULL);
	return FALSE;
}

static int cmd_uldashd (Word *w, int align, char has_param, short param) {
	attr_push(ATTR_DOT_DASH_UL,NULL);
	return FALSE;
}

static int cmd_uldashdd (Word *w, int align, char has_param, short param) {
	attr_push(ATTR_2DOT_DASH_UL,NULL);
	return FALSE;
}

static int cmd_ulw (Word *w, int align, char has_param, short param) {
	attr_push(ATTR_WORD_UL,NULL);
	return FALSE;
}

static int cmd_ulth (Word *w, int align, char has_param, short param) {
	attr_push(ATTR_THICK_UL,NULL);
	return FALSE;
}

static int cmd_ulwave (Word *w, int align, char has_param, short param) {
	attr_push(ATTR_WAVE_UL,NULL);
	return FALSE;
}

static int cmd_strike (Word *w, int align, char has_param, short param) {
	if (has_param && param==0)
		attr_pop(ATTR_STRIKE);
	else
		attr_push(ATTR_STRIKE,NULL);
	return FALSE;
}
static int cmd_strikedl (Word *w, int align, char has_param, short param) {
	if (has_param && param==0)
		attr_pop(ATTR_DBL_STRIKE);
	else
		attr_push(ATTR_DBL_STRIKE,NULL);
	return FALSE;
}
static int cmd_striked (Word *w, int align, char has_param, short param) {
	if (has_param && param==0)
		attr_pop(ATTR_DBL_STRIKE);
	else
		attr_push(ATTR_DBL_STRIKE,NULL);
	return FALSE;
}


/*-------------------------------------------------------------------*/

static int cmd_rtf (Word *w, int align, char has_param, short param) {
#if 0
	if (!inline_mode)
		printf ("<html>\n<head>\n");
	print_banner();
#endif
	return FALSE;
}

static int cmd_up (Word *w, int align, char has_param, short param) {
	if (has_param && param==0)
		attr_pop(ATTR_SUPER);
	else
		attr_push(ATTR_SUPER,NULL);
	return FALSE;
}

static int cmd_dn (Word *w, int align, char has_param, short param) {
	if (has_param && param==0)
		attr_pop(ATTR_SUB);
	else
		attr_push(ATTR_SUB,NULL);
	return FALSE;
}

static int cmd_shad (Word *w, int align, char has_param, short param) {
	if (has_param && param==0)
		attr_pop(ATTR_SHADOW);
	else
		attr_push(ATTR_SHADOW,NULL);
	return FALSE;
}

static int cmd_b (Word *w, int align, char has_param, short param) {
	if (has_param && param==0)
		attr_pop(ATTR_BOLD);
	else
		attr_push(ATTR_BOLD,NULL);
	return FALSE;
}

static int cmd_i (Word *w, int align, char has_param, short param) {
	if (has_param && param==0)
		attr_pop(ATTR_ITALIC);
	else
		attr_push(ATTR_ITALIC,NULL);
	return FALSE;
}

static int cmd_s (Word *w, int align, char has_param, short param) {
	return FALSE;
}

static int cmd_outl (Word *w, int align, char has_param, short param) {
	if (has_param && param==0)
		attr_pop(ATTR_OUTLINE);
	else
		attr_push(ATTR_OUTLINE,NULL);
	return FALSE;
}

static int cmd_ansi (Word *w, int align, char has_param, short param) {
	charset_type = CHARSET_ANSI;
	printf ("<!-- document uses ANSI character set -->\n");
	return FALSE;
}

static int cmd_pc (Word *w, int align, char has_param, short param) {
	charset_type = CHARSET_CP437 ;
	printf ("<!-- document uses PC codepage 437 character set -->\n");
	return FALSE;
}

static int cmd_pca (Word *w, int align, char has_param, short param) {
	charset_type = CHARSET_CP850;
	printf ("<!-- document uses PC codepage 850 character set -->\n");
	return FALSE;
}

static int cmd_mac (Word *w, int align, char has_param, short param) {
	charset_type = CHARSET_MAC;
	printf ("<!-- document uses Macintosh character set -->\n");
	return FALSE;
}

static int cmd_colortbl (Word *w, int align, char has_param, short param) {
	if (w->next) {
		process_color_table(w->next);
	}
	return TRUE;
}

static int cmd_fonttbl (Word *w, int align, char has_param, short param) {
	if (w->next) {
		process_font_table(w->next);
	}
	return TRUE;
}

static int cmd_header (Word *w, int align, char has_param, short param) {
	return TRUE;
}

static int cmd_headerl (Word *w, int align, char has_param, short param) {
	return TRUE;
}

static int cmd_headerr (Word *w, int align, char has_param, short param) {
	return TRUE;
}

static int cmd_headerf (Word *w, int align, char has_param, short param) {
	return TRUE;
}

static int cmd_footer (Word *w, int align, char has_param, short param) {
	return TRUE;
}

static int cmd_footerl (Word *w, int align, char has_param, short param) {
	return TRUE;
}

static int cmd_footerr (Word *w, int align, char has_param, short param) {
	return TRUE;
}

static int cmd_footerf (Word *w, int align, char has_param, short param) {
	return TRUE;
}

/* dummy function to get rid of subgroups */
static int cmd_ignore (Word *w, int align, char has_param, short param) {
	return TRUE;
}

static int cmd_info (Word *w, int align, char has_param, short param) {
	process_info_group (w->next);
	return TRUE;
}

/*----\pict command-------------------------------------------------------*/
static int cmd_pict (Word *w, int align, char has_param, short param) {
	within_picture=TRUE;
	picture_width = picture_height = 0;
	picture_type = PICT_WB;
	return FALSE;		
}
static int cmd_bin (Word *w, int align, char has_param, short param) {
	return FALSE;		
}
static int cmd_macpict (Word *w, int align, char has_param, short param) {
	picture_type = PICT_MAC;
	return FALSE;		
}
static int cmd_pnmetafile (Word *w, int align, char has_param, short param) {
	picture_type = PICT_PM;
	return FALSE;		
}
static int cmd_wmetafile (Word *w, int align, char has_param, short param) {
	picture_type = PICT_WM;
	if (within_picture && has_param) {
		picture_wmetafile_type=param;
		switch(param) {
		case 1: picture_wmetafile_type_str="MM_TEXT"; break;
		case 2: picture_wmetafile_type_str="MM_LOMETRIC"; break;
		case 3: picture_wmetafile_type_str="MM_HIMETRIC"; break;
		case 4: picture_wmetafile_type_str="MM_LOENGLISH"; break;
		case 5: picture_wmetafile_type_str="MM_HIENGLISH"; break;
		case 6: picture_wmetafile_type_str="MM_TWIPS"; break;
		case 7: picture_wmetafile_type_str="MM_ISOTROPIC"; break;
		case 8: picture_wmetafile_type_str="MM_ANISOTROPIC"; break;
		default: picture_wmetafile_type_str="default:MM_TEXT"; break;
		}
	}
	return FALSE;		
}

static int cmd_wbmbitspixel (Word *w, int align, char has_param, short param) {
	if (within_picture && has_param) 
		picture_bits_per_pixel = param;
	return FALSE;		
}

static int cmd_picw (Word *w, int align, char has_param, short param) {
	if (within_picture && has_param) 
		picture_width = param;
	return FALSE;		
}
static int cmd_pich (Word *w, int align, char has_param, short param) {
	if (within_picture && has_param) 
		picture_height = param;
	return FALSE;		
}


typedef struct {
	char *name;
	int (*func)(Word*,int,char,short);
	char *always_print;
	char *debug_print;
}
HashItem;



static HashItem hash_other [] = {
	{ "*", cmd_ignore, NULL, NULL },
	{ "-", NULL, NULL, "optional hyphen" },
	{ "_", NULL, "-", "nonbreaking hyphen" },
	{ "~", NULL, "&nbsp;", NULL },
	{ NULL, NULL, NULL, NULL}
};
static HashItem hash_a [] = {
	{ "ansi", &cmd_ansi , NULL, NULL },
	{ NULL, NULL, NULL, NULL}
};
static HashItem hash_b [] = {
	{ "b", &cmd_b, NULL, NULL },
	{ "bullet", &cmd_bullet, NULL, NULL },
	{ "bin", &cmd_bin, NULL, "picture is binary" },
#if 0
	{ "bgbdiag", NULL, NULL, NULL },
	{ "bgcross", NULL, NULL, NULL },
	{ "bgdcross", NULL, NULL, NULL },
	{ "bgfdiag", NULL, NULL, NULL },
	{ "bghoriz", NULL, NULL, NULL },
	{ "bgkbdiag", NULL, NULL, NULL },
	{ "bgkcross", NULL, NULL, NULL },
	{ "bgkdcross", NULL, NULL, NULL },
	{ "bgkfdiag", NULL, NULL, NULL },
	{ "bgkhoriz", NULL, NULL, NULL },
	{ "bgkvert", NULL, NULL, NULL },
	{ "bgvert", NULL, NULL, NULL },
	{ "brdrcf", NULL, NULL, NULL },
	{ "brdrdb", NULL, NULL, NULL },
	{ "brdrdot", NULL, NULL, NULL },
	{ "brdrhair", NULL, NULL, NULL },
	{ "brdrs", NULL, NULL, NULL },
	{ "brdrsh", NULL, NULL, NULL },
	{ "brdrth", NULL, NULL, NULL },
	{ "brdrw", NULL, NULL, NULL },
#endif
	{ NULL, NULL, NULL, NULL}
};
static HashItem hash_c [] = {
	{ "caps", &cmd_caps, NULL, NULL },
	{ "cb", cmd_cb, NULL, NULL },
	{ "cf", cmd_cf, NULL, NULL },
	{ "colortbl", &cmd_colortbl, NULL, "color table" },
	{ "cols", NULL, NULL, "columns (not implemented)" },
	{ "column", NULL, NULL, "column break (not implemented)" },
#if 0
	{ "cbpat", NULL, NULL, NULL },
	{ "cellx", NULL , NULL, NULL },
	{ "cfpat", NULL, NULL, NULL },
	{ "cgrid", NULL , NULL, NULL },
	{ "clbgbcross", NULL , NULL, NULL },
	{ "clbgbdiag", NULL , NULL, NULL },
	{ "clbgbkbdiag", NULL , NULL, NULL },
	{ "clbgbkcross", NULL , NULL, NULL },
	{ "clbgbkdcross", NULL , NULL, NULL },
	{ "clbgbkfdiag", NULL , NULL, NULL },
	{ "clbgbkhor", NULL , NULL, NULL },
	{ "clbgbkvert", NULL , NULL, NULL },
	{ "clbgdcross", NULL , NULL, NULL },
	{ "clbgfdiag", NULL , NULL, NULL },
	{ "clbghoriz", NULL , NULL, NULL },
	{ "clbgvert", NULL , NULL, NULL },
	{ "clbrdrb", NULL , NULL, NULL },
	{ "clbrdrl", NULL , NULL, NULL },
	{ "clbrdrr", NULL , NULL, NULL },
	{ "clbrdrt", NULL , NULL, NULL },
	{ "clcbpat", NULL , NULL, NULL },
	{ "clcfpat", NULL , NULL, NULL },
	{ "clmgf", NULL , NULL, NULL },
	{ "clmrg", NULL , NULL, NULL },
	{ "clshdng", NULL , NULL, NULL },
#endif
	{ NULL, NULL, NULL, NULL}
};
static HashItem hash_d [] = {
	{ "dn", &cmd_dn, NULL, NULL },
#if 0
	{ "dibitmap", NULL, NULL, NULL },
#endif
	{ NULL, NULL, NULL, NULL}
};
static HashItem hash_e [] = {
	{ "emdash", NULL, "&mdash;", NULL },
	{ "endash", NULL, "&ndash;", NULL },
	{ "expnd", &cmd_expand, NULL, NULL },
	{ "embo", &cmd_emboss, NULL, NULL },
	{ "expand", &cmd_expand, NULL, NULL },
	{ NULL, NULL, NULL, NULL}
};
static HashItem hash_f [] = {
	{ "f", cmd_f, NULL, NULL },
	{ "fdecor", cmd_fdecor, NULL, NULL },
	{ "fmodern", cmd_fmodern, NULL, NULL },
	{ "fnil", cmd_fnil, NULL, NULL },
	{ "fonttbl", cmd_fonttbl, NULL, "font table" },
	{ "froman", cmd_froman, NULL, NULL },
	{ "fs", cmd_fs, NULL, NULL },
	{ "fscript", cmd_fscript, NULL, NULL },
	{ "fswiss", cmd_fswiss, NULL, NULL },
	{ "ftech", cmd_ftech, NULL, NULL },
	{ "field", cmd_field, NULL, NULL },
	{ "footer", cmd_footer, NULL, NULL },
	{ "footerf", cmd_footerf, NULL, NULL },
	{ "footerl", cmd_footerl, NULL, NULL },
	{ "footerr", cmd_footerr, NULL, NULL },
	{ NULL, NULL, NULL, NULL}
};
static HashItem hash_h [] = {
	{ "highlight", &cmd_highlight, NULL, NULL },
	{ "header", cmd_header, NULL, NULL },
	{ "headerf", cmd_headerf, NULL, NULL },
	{ "headerl", cmd_headerl, NULL, NULL },
	{ "headerr", cmd_headerr, NULL, NULL },
	{ "hl", cmd_ignore, NULL, "hyperlink within object" },
	{ NULL, NULL, NULL, NULL}
};
static HashItem hash_i [] = {
	{ "i", &cmd_i, NULL, NULL },
	{ "info", &cmd_info, NULL, NULL },
	{ "intbl", &cmd_intbl, NULL, NULL },
	{ "impr", &cmd_engrave, NULL, NULL },
	{ NULL, NULL, NULL, NULL}
};
static HashItem hash_l [] = {
	{ "ldblquote", &cmd_ldblquote, NULL, NULL },
	{ "line", &cmd_line, NULL, NULL },
	{ "lquote", &cmd_lquote, NULL, NULL },
	{ NULL, NULL, NULL, NULL}
};
static HashItem hash_m [] = {
	{ "mac", &cmd_mac , NULL, NULL },
	{ "macpict", &cmd_macpict, NULL, NULL },
	{ NULL, NULL, NULL, NULL}
};
static HashItem hash_o [] = {
	{ "outl", &cmd_outl, NULL, NULL },
	{ NULL, NULL, NULL, NULL}
};
static HashItem hash_p [] = {
	{ "page", &cmd_page, NULL, NULL },
	{ "par", &cmd_par, NULL, NULL },
	{ "pc", &cmd_pc , NULL, NULL },
	{ "pca", &cmd_pca , NULL, NULL },
	{ "pich", &cmd_pich, NULL, NULL },
	{ "pict", &cmd_pict, NULL, "picture" },
	{ "picw", &cmd_picw, NULL, NULL },
	{ "plain", &cmd_plain, NULL, NULL },
	{ "pnmetafile", &cmd_pnmetafile, NULL, NULL },
#if 0
	{ "piccropb", NULL, NULL, NULL },
	{ "piccropl", NULL, NULL, NULL },
	{ "piccropr", NULL, NULL, NULL },
	{ "piccropt", NULL, NULL, NULL },
	{ "pichgoal", NULL, NULL, NULL },
	{ "pichgoal", NULL, NULL, NULL },
	{ "picscaled", NULL, NULL, NULL },
	{ "picscalex", NULL, NULL, NULL },
	{ "picwgoal", NULL, NULL, NULL },
#endif
	{ NULL, NULL, NULL, NULL}
};
static HashItem hash_r [] = {
	{ "rdblquote", &cmd_rdblquote, NULL, NULL },
	{ "rquote", &cmd_rquote, NULL, NULL },
	{ "rtf", &cmd_rtf, NULL, NULL },
	{ NULL, NULL, NULL, NULL}
};
static HashItem hash_s [] = {
	{ "s", cmd_s, NULL, "style" },
	{ "sect", NULL, "<p>", "section break"},
	{ "scaps", &cmd_scaps, NULL, NULL },
	{ "shad", &cmd_shad, NULL, NULL },
	{ "strike", &cmd_strike, NULL, NULL },
	{ "striked", &cmd_striked, NULL, NULL },
	{ "strikedl", &cmd_strikedl, NULL, NULL },
	{ "stylesheet", &cmd_ignore, NULL, "style sheet" },
#if 0
	{ "shading", NULL, NULL, NULL },
#endif
	{ NULL, NULL, NULL, NULL}
};
static HashItem hash_t [] = {
	{ "tab", &cmd_tab, NULL, NULL },
	{ "tc", cmd_ignore, NULL, "TOC entry" },
	{ "tcn", cmd_ignore, NULL, "TOC entry" },
#if 0
	{ "tcf", NULL , NULL, NULL },
	{ "tcl", NULL , NULL, NULL },
	{ "trgaph", NULL , NULL, NULL },
	{ "trleft", NULL , NULL, NULL },
	{ "trowd", NULL , NULL, NULL },
	{ "trqc", NULL , NULL, NULL },
	{ "trql", NULL , NULL, NULL },
	{ "trqr", NULL , NULL, NULL },
	{ "trrh", NULL , NULL, NULL },
#endif
	{ NULL, NULL, NULL, NULL}
};
static HashItem hash_u [] = {
	{ "ul", &cmd_ul, NULL, NULL },
	{ "up", &cmd_up, NULL, NULL },
	{ "uld", &cmd_uld, NULL, NULL },
	{ "uldash", &cmd_uldash, NULL, NULL },
	{ "uldashd", &cmd_uldashd, NULL, NULL },
	{ "uldashdd", &cmd_uldashdd, NULL, NULL },
	{ "uldb", &cmd_uldb, NULL, NULL },
	{ "ulnone", &cmd_ulnone, NULL, NULL },
	{ "ulth", &cmd_ulth, NULL, NULL },
	{ "ulw", &cmd_ulw, NULL, NULL },
	{ "ulwave", &cmd_ulwave, NULL, NULL },
	{ NULL, NULL, NULL, NULL}
};

static HashItem hash_w [] = {
	{ "wbmbitspixel", &cmd_wbmbitspixel, NULL, NULL },
	{ "wmetafile", &cmd_wmetafile, NULL, NULL },
#if 0
	{ "wbitmap", NULL, NULL, NULL },
	{ "wbmplanes", NULL, NULL, NULL },
	{ "wbmwidthbytes", NULL, NULL, NULL },
#endif
	{ NULL, NULL, NULL, NULL}
};

static HashItem hash_x [] = {
	{ "xe", cmd_ignore, NULL, "index entry" },
	{ NULL, NULL, NULL, NULL}
};

static HashItem *hashes [26] = {
	hash_a,
	hash_b,
	hash_c,
	hash_d,
	hash_e,
	hash_f,
	NULL,
	hash_h,
	hash_i,
	NULL,
	NULL,
	hash_l,
	hash_m,
	NULL,
	hash_o,
	hash_p,
	NULL,
	hash_r,
	hash_s,
	hash_t,
	hash_u,
	NULL,
	hash_w,
	hash_x, 
	NULL, NULL
};


/*-------------------------------------------------------------------*/
/*-------------------------------------------------------------------*/



/*-------------------------------------------------------------------*/
/*-------------------------------------------------------------------*/

void print_with_special_exprs (char *s) {
	int ch;
	int state;

enum { SMALL=0, BIG=1 };

	CHECK_PARAM_NOT_NULL(s);

	if (simulate_smallcaps) {
		if (*s >= 'a' && *s <= 'z') {
			state=SMALL;
			printf ("<small>");
		}
		else
			state=BIG;
	}

	while ((ch=*s)) {
		if (simulate_allcaps || simulate_smallcaps)
			ch = toupper (ch);

		if (ch=='<') printf ("&lt;");
		else if (ch=='>') printf ("&gt;");
		else if (ch=='&') printf ("&amp;");
		else printf ("%c",ch);

		s++;

		if (simulate_smallcaps) {
			ch = *s;
			if (ch >= 'a' && ch <= 'z') {
				if (state==BIG)
					printf ("<small>");
				state=SMALL;
			}
			else
			{
				if (state==SMALL)
					printf ("</small>");
				state=BIG;
			}
		}
	}
}


static void begin_table()
{
	within_table=TRUE;
	have_printed_tr = FALSE;
	have_printed_td = FALSE;
	have_printed_end_tr = FALSE;
	have_printed_end_td = FALSE;
	attr_push_mark();
	starting_body();
	printf ("<table border=1>\n");
}


void end_table ()
{
	if (within_table) {
		if (!have_printed_end_td) {
			attr_pop_dump();
			printf ("</td>");
		}
		if (!have_printed_end_tr) {
			printf ("</tr>\n");
		}
		printf ("</table>\n");
		within_table=FALSE;
		have_printed_tr = FALSE;
		have_printed_td = FALSE;
		have_printed_end_tr = FALSE;
		have_printed_end_td = FALSE;
	}
}


void starting_text() {
	if (within_table) {
		if (!have_printed_tr) {
			printf ("<tr>");
			have_printed_tr=TRUE;
			have_printed_end_tr=FALSE;
			have_printed_td=FALSE;
		}
		if (!have_printed_td) {
			printf ("<td>");
			attr_push_dump();
			have_printed_td=TRUE;
			have_printed_end_td=FALSE;
		}
	}
}

static void
starting_paragraph_align (int align)
{
	if (within_head && align != ALIGN_LEFT)
		starting_body();

	switch (align) 
	{
	case ALIGN_CENTER:
		printf ("<center>"); break;
	case ALIGN_LEFT:
		break;
	case ALIGN_RIGHT:
		if (simple_mode)
			printf ("<div align=right>\n");
		else
			printf ("<div style=\"text-align:right\">\n"); 
		break;
	case ALIGN_JUSTIFY:
		if (simple_mode)
			printf ("<div align=justify>\n"); 
		else
			printf ("<div style=\"text-align:justify\">\n"); 
		break;
	}
}


static void
ending_paragraph_align (int align)
{
	switch (align) {
	case ALIGN_CENTER:
		printf ("</center>"); break;
	case ALIGN_LEFT:
		break;
	case ALIGN_RIGHT:
	case ALIGN_JUSTIFY:
		printf ("</div>\n"); 
		break;
	}
}

/* Recursive routine to print out HTML based on a tree of words.
 */
static void
word_print_html_core (Word *w)
{
	char *s;
	FILE *f=NULL;
	int is_cell_group=FALSE;
	int paragraph_started=FALSE;
	int paragraph_align=ALIGN_LEFT;

	CHECK_PARAM_NOT_NULL(w);

	if (!coming_pars_that_are_tabular && within_table) {
		end_table();
	}
	else if (coming_pars_that_are_tabular && !within_table) {
		begin_table();
	}

	/* Mark our place in the stack */
	attr_push_mark();

	while (w) {
		s = word_string (w);

#if 0
		if (debug_mode) {
			printf ("<!-- word='%s' -->\n", s);
		}
#endif

		if (s) {
			/*--Handle word -----------------------------------*/

			if (s[0] != '\\') {
				starting_body();

#if 0
				if (debug_mode)
				  printf ("<!-- have_printed_tr=%d --->\n", have_printed_tr);
				if (debug_mode)
				  printf ("<!-- have_printed_td=%d --->\n", have_printed_td);
#endif
				starting_text();

				if (!paragraph_started) {
					starting_paragraph_align (paragraph_align);
					paragraph_started=TRUE;
				}

				/*----------------------------------------*/
				if (within_picture) {
					starting_body();
					if (!f) {
						char *ext=NULL;
						switch (picture_type) {
						case PICT_WB: ext="bmp"; break;
						case PICT_WM: ext="wmf"; break;
						case PICT_MAC: ext="pict"; break;
						case PICT_DI: ext="dib"; break; /* Device independent bitmap=??? */
						case PICT_PM: ext="pmm"; break; /* OS/2 metafile=??? */
						}
						sprintf (picture_path, "pict%03d.%s", 
							picture_file_number++,ext);
						f=fopen(picture_path,"w");
					}

					if (s[0]!=' ') {
						char *s2;
						printf ("<!---picture data found--->\n");
						if (picture_wmetafile_type_str) {
							printf ("<!---Windows Metafile type is %s--->\n",
								picture_wmetafile_type_str);
						}
						printf ("<!---picture dimensions: %d by %d, depth %d--->\n", 
							picture_width, picture_height, picture_bits_per_pixel);
						if (picture_width && picture_height && picture_bits_per_pixel) {
							s2=s;
							while (*s2) {
								unsigned int tmp,value;
								tmp=tolower(*s2++);
								if (tmp>'9') tmp-=('a'-10); 
								else tmp-='0';
								value=16*tmp;
								tmp=tolower(*s2++);
								if (tmp>'9') tmp-=('a'-10); 
								else tmp-='0';
								value+=tmp;
								fprintf (f,"%c", value);
							}
						}
					}
				}
				/*----------------------------------------*/
				else {
					total_chars_this_line += strlen(s);
					print_with_special_exprs (s);
				}

			/*---Handle RTF keywords---------------------------*/

			} else {
				int done=FALSE;

				s++;

/*----Paragraph alignment----------------------------------------------------*/
				if (!strcmp ("ql", s))
					paragraph_align = ALIGN_LEFT;
				else if (!strcmp ("qr", s))
					paragraph_align = ALIGN_RIGHT;
				else if (!strcmp ("qj", s))
					paragraph_align = ALIGN_JUSTIFY;
				else if (!strcmp ("qc", s))
					paragraph_align = ALIGN_CENTER;
				else if (!strcmp ("pard", s)) 
				{
					/* Clear out all font attributes.
					 */
					attr_pop_all();
					if(coming_pars_that_are_tabular) {
						--coming_pars_that_are_tabular;
					}

					/* Clear out all paragraph attributes.
					 */
					ending_paragraph_align(paragraph_align);
					paragraph_align = ALIGN_LEFT;
					paragraph_started = FALSE;
				}
/*----Table keywords---------------------------------------------------------*/
				else
				if (!strcmp (s, "cell")) {
					is_cell_group=TRUE;
					if (!have_printed_td) {
						/* Need this with empty cells */
						printf ("<td>");
						attr_push_dump();
					}
					attr_pop_dump();
					printf ("</td>");
					have_printed_td = FALSE;
					have_printed_end_td=TRUE;
				}
				else if (!strcmp (s, "row")) {
					if (within_table) {
						printf ("</tr>\n");
						have_printed_tr = FALSE;
						have_printed_end_tr=TRUE;
					} else {
						if (debug_mode)
						  printf ("<!-- end of table row -->\n");
					}
				}

/*----Special chars---------------------------------------------------------*/
				else if (*s == '\'') {
					/* \'XX is a hex char code expression */
					int ch = h2toi (&s[1]);
					if (ch >= 32 && ch < 127) {
						putchar (ch);
					} if (ch >= 160) {
						char *s2;
						if (ch==0xa9) {
							printf ("&copy;");
							++total_chars_this_line; /* \tab */
						} else if (ch==0xae) {
							printf ("&reg;");
							++total_chars_this_line; /* \tab */
						} else {
							s2 = special_exprs_lookup (charset_type, ch);
							if (!s2 || !*s2) 
								printf("<!--0x%02x-->",ch);
							else
								printf ("&%s;", s2);
						}
					}
				}
				else
/*----Search the RTF command hash-------------------------------------------*/
				{
					int ch;
					int index=0;
					int have_param=FALSE, param=0;
					HashItem *hip;
					char *p;

#if 0
					if (debug_mode)
					  printf ("<!-- interpreting command \\%s -->\n", s);
#endif

					/* Look for a parameter */
					p=s;
					while(*p && (!isdigit(*p) && *p!='-')) p++;
					if (*p && (isdigit(*p) || *p=='-')) { 
						have_param=TRUE; 
						param=atoi (p);
					}

#if 0
					if (debug_mode) {
					  if (have_param)
					   printf ("<!-- param = %d -->\n", param);
					  else 
					   printf ("<!-- no param -->\n");
					}
#endif

					/* Generate a hash index
					 */
					ch = tolower (*s);
					if (ch>='a' && ch<='z') 
						hip = hashes[ch-'a'];
					else
						hip = hash_other;

					if (!hip) {
						if (debug_mode) 
							printf ("<!-- unfamiliar rtf command: %s -->\n", s);
					}
					else
					while (!done) {
						int match=FALSE;

						if (have_param)
							match = !strncmp (s, hip[index].name, p-s);
						else
							match = !strcmp(s, hip[index].name);

						if (match) {
							char *always;
							char *debug;
							int terminate_group;

							if (hip[index].func) {
								terminate_group = hip[index].func (w,paragraph_align, have_param, param);

								if (terminate_group)
									while(w) w=w->next;
							}

							always=hip[index].always_print;
							debug=hip[index].debug_print;

							if (always)
								printf ("%s", always);
							if (debug && debug_mode)
								printf ("<!-- %s -->", debug);

							done=TRUE;
						}
						else
						{
							index++;
							if (!hip[index].name)
								done=TRUE;
						}
					}
				}
			}
/*-------------------------------------------------------------------------*/
		} else {
			Word *child = w->child;

			if (!paragraph_started) {
				starting_paragraph_align (paragraph_align);
				paragraph_started=TRUE;
			}

			if (child)
			  word_print_html_core (child);
		}

		if (w) 
			w = w->next;
	}

	if (within_picture) {
		if(f) { 
			fclose(f);
			printf ("<img src=%s>\n", picture_path);
			printf ("<!---end of picture--->\n");
		}
	}

	/* Undo font attributes UNLESS we're doing table cells
	 * since they would appear between </td> and </tr>.
	 */
	if (!is_cell_group)
		attr_pop_all();
	else
		attr_drop_all();

	/* Undo paragraph alignment
	 */
	if (paragraph_started)
		ending_paragraph_align (paragraph_align);

	attr_pop_mark();
}



void word_print_html (Word *w) 
{
	CHECK_PARAM_NOT_NULL (w);

	if (!inline_mode)
		printf ("<HTML><HEAD>\n");

	print_banner ();

	within_head=TRUE;
	have_printed_body=FALSE;
	within_table=FALSE;
	simulate_allcaps=FALSE;
	word_print_html_core (w);
	end_table();

	if (!inline_mode)
		printf ("\n</BODY></HTML>\n");
}


