
/*=============================================================================
   rtf2htm, a command-line program to convert RTF-formatted documents to HTML.
   Copyright (C) 2000,2001 Zachary Thayer Smith

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   The author is reachable by electronic mail at tuorfa@yahoo.com.
=============================================================================*/


/*----------------------------------------------------------------------
 * Module name:    parse
 * Author name:    Zach Smith
 * Create date:    01 Sep 2000
 * Last change:    03 Aug 2001
 * Purpose:        Parsing of the RTF file into a structure of Word objects.
 *----------------------------------------------------------------------
 * Changes:
 * 15 Oct 00, tuorfa@yahoo.com: parse.c created with functions taken from word.c
 * 15 Oct 00, tuorfa@yahoo.com: backslash before newline is now \par
 * 08 Apr 01, tuorfa@yahoo.com: removed limit on word length
 * 03 Aug 01, tuorfa@yahoo.com: added input buffering
 *--------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#include "defs.h"
#include "parse.h"
#include "malloc.h"
#include "main.h"
#include "error.h"
#include "word.h"
#include "hash.h"


/* local to word_dump */
static int indent_level=-1;

/* local to getchar stuff */
static int ungot_char=-1;
static int ungot_char2=-1;
static int ungot_char3=-1;


static void my_unget_char (int ch)
{
	if (ungot_char>=0 && ungot_char2>=0 && ungot_char3>=0) 
		error_handler ("more than 3 ungot chars");

	ungot_char3 = ungot_char2;
	ungot_char2 = ungot_char;
	ungot_char = ch;
}

static int last_returned_ch=0;


#define READ_BUF_LEN 4096
static int buffer_size = 0;
static char *read_buf = NULL;
static int read_buf_end = 0;
static int read_buf_index = 0;


static int my_getchar (FILE* f)
{
	int ch;

	CHECK_PARAM_NOT_NULL(f);

	if (ungot_char>=0) {
		ch = ungot_char; 
		ungot_char=ungot_char2; 
		ungot_char2=ungot_char3;
		ungot_char3=-1;
		last_returned_ch = ch;
		return ch;
	}
	do {
		if (read_buf_index >= read_buf_end) {
			if (!read_buf) {
				buffer_size = READ_BUF_LEN;
				read_buf = my_malloc (buffer_size);
				if (!read_buf) {
					buffer_size /= 4;
					read_buf = my_malloc (buffer_size);
					if (!read_buf) 
						error_handler ("cannot allocate read buffer");
				}
			}
			read_buf_end = fread (read_buf, 1, buffer_size, f);
			read_buf_index = 0;
			if (!read_buf_end)
				return EOF;
		}
		ch = read_buf [read_buf_index++];

		if (ch=='\n') {
			lineno++;
			/* Convert \(newline) into \par here */
			if (last_returned_ch=='\\') {
				my_unget_char (' ');
				my_unget_char ('r');
				my_unget_char ('a');
				ch = 'p';
				break;
			}
		}
	} 
	while (ch=='\r' || ch=='\n');

	if (ch=='\t') ch = ' ';

	last_returned_ch = ch;
	return ch;
}


/* local to read_word */
static char *input_str = NULL;
static unsigned long current_max_length = 1;

int
expand_word_buffer ()
{
	char *new_ptr;
	unsigned long old_length;
	if (!input_str)
		error_handler ("no input buffer allocated");
	old_length = current_max_length;
	current_max_length *= 2;
	new_ptr = my_malloc (current_max_length);
	if (new_ptr) 
	{
		memcpy (new_ptr, input_str, old_length);
		my_free (input_str);
		input_str = new_ptr;
		return TRUE;
	} 
	return FALSE;
}


static int 
read_word (FILE *f) 
{
	int ch, ix=0;
	int have_whitespace=FALSE;
	int is_control_word=FALSE;
	int has_numeric_param=FALSE; /* if is_control_word==TRUE */

	CHECK_PARAM_NOT_NULL(f);

	current_max_length = 10; /* XX */

	/* Get some storage for a word.
	 */
	input_str = my_malloc (current_max_length);
	if (!input_str)
		error_handler("cannot allocate word storage");

	ch = my_getchar(f);
	while (ch == ' ') {
		ch = my_getchar(f);
		have_whitespace=TRUE;
	}
	if (ch==EOF) return 0;

	if (have_whitespace) {
		my_unget_char (ch);
		input_str[0]=' '; 
		input_str[1]=0;
		return 1;
	}

	/* Take care of single-character tokens */
	if (ch == '{' || ch == '}' || ch == ';') {
		input_str[0]=ch; input_str[1]=0;
		return 1;
	}

	/* We need to know if this is a control word,
	 * since they can terminate differently than text words.
	 */
	if (ch=='\\') {
		is_control_word=TRUE;
		ix=1;
		input_str[0]=ch;
		ch=my_getchar(f);
	}

	while (ch!=EOF && ch!=' ')
	{
		/* In RTF, a tab char is the same as \tab.
		 */
		if (!ix && ch=='\t') {
			strcpy (input_str, "\\tab");
			ix=4;
			break;
		}

		/* Newlines mean nothing in RTF files unless
		 * preceded by a backslash, when they mean \par.
		 */
		if (ch=='\n') { 
			if (ix==1 && input_str[0]=='\\') {
				strcpy (input_str, "\\par");
				ix=4;
				break;
			}
			ch = my_getchar(f); 
			continue; 
		}

		/* Take care of RTF's various special two-characters 
		 * expressions.
		 */
		if (ix==1 && input_str[0]=='\\') {
			if (ch=='{' || ch=='}' || ch=='\\' || ch=='~'
			    || ch=='_' || ch=='-')
			{
				input_str[ix++] = ch;
				break;
			}
		}
		
		/* Preserve \'## expressions (hex char exprs) for later.
		 */
		if (ix==1 && ch == '\'' && input_str[0]=='\\') {
			input_str[ix++]=ch;
			if(ix==current_max_length) {
				if (!expand_word_buffer ())
					error_handler("word too long");
			}
			ch = my_getchar(f);
			input_str[ix++]=ch;
			if(ix==current_max_length) {
				if (!expand_word_buffer ())
					error_handler("word too long");
			}
			ch = my_getchar(f);
			input_str[ix++]=ch;
			if(ix==current_max_length) {
				if (!expand_word_buffer ())
					error_handler("word too long");
			}
			break;
		}

		/* Terminate a control word */
		if (is_control_word) {
			if (ch==' ' || ch==';')
				break;

			/* In RTF, a control word can terminate if its
			 * numeric parameter is immediately followed by
			 * a non-space character.
			 */
			if (has_numeric_param && !isdigit(ch)) {
				my_unget_char(ch);
				break;
			}
			else if (isdigit(ch))
				has_numeric_param=TRUE;
		}
		
		/* { and } and ; are returned as separate words */
		if (!ix && (ch == '}' || ch == '{' || ch == ';')) {
			input_str[ix++] = ch; 
			break;
		}

		/* Take care of characters which can only begin words. */
		if ((ch=='{' || ch=='}' || ch=='\t' || ch == '\\') && ix) { 
			my_unget_char (ch); 
			break; 
		}

		input_str[ix++] = ch;
		if(ix==current_max_length) {
			if (!expand_word_buffer ())
				error_handler("word too long");
		}
		ch = my_getchar (f);
	}

	/* Retain the trailing space char, unless this is a
	 * control word.
	 */
	if (ch==' ' && input_str[0]!='\\')
		my_unget_char(ch);

	/* Retain trailing semicolon if it is a control word.
	 */
	if (ch==';' && input_str[0]=='\\')
		my_unget_char(ch);

	input_str[ix]=0;
	return ix;
}


/* Function word_read is the metareader which pieces together the structure of
 * Word objects.
 */
Word *
word_read (FILE* f) {
	Word * prev_word = NULL;
	Word * first_word = NULL;
	Word * new_word = NULL; /* temp */

	CHECK_PARAM_NOT_NULL(f);

	do {
		if (!read_word(f)) {
			return first_word;
		}

		if (input_str[0] == '{') {
			/* Process subwords */

#if 0
printf ("processing subword...\n");
#endif

			/* Create a dummy word to point to a sublist */
			new_word = word_new (NULL);
			if (!new_word)
				error_handler ("cannot allocate word");

			/* Get the sublist */
			new_word->child = word_read (f);
			if (!new_word->hash_index && !new_word->child)
			{
				printf ("unable to read children!\n");
			}

		} else if (input_str[0] == '}') {
#if 0
printf ("returning from word_read.\n");
#endif
			return first_word;
		} else {
			new_word = word_new (input_str);
		}

		if (prev_word) prev_word->next = new_word;

		if (!first_word) first_word = new_word;

		prev_word = new_word;

		/* Free up the memory allocated by read_word.
		 */
		my_free (input_str);
		input_str = NULL;
	}
	while(1);

}


/* Recursive diagnostic routine to print out a tree of words.
 */
void
word_dump (Word *w)
{
	int i;
	char *s;

	CHECK_PARAM_NOT_NULL(w);

	printf ("\n");
	indent_level += 2;
	if (indent_level) {
		for (i=0;i<indent_level;i+=2)
			printf ("| ");
	} else {
		printf ("\n-----------------------------------------------------------------------\n\n");
	}

	while (w) {
		s = word_string (w);
		if (s) { 
			printf ("\"%s\" ", s);
		} else { 
			if (w->child) 
				word_dump (w->child); 
			else
				warning_handler ("Word object has no string and no children");
		}
		w = w->next;
	}

	indent_level -= 2;
}


